;;; Extra wonderful routines taken from TMC and the Net
;;; Search for "Table of contents" for a list.

;;; Copyright (C) Schlumberger Cambridge Research, 1989
;;; subject to the terms of the GNU general public licence

;;; library file for miscellaneous interactive functions

(defun backward-kill-line ()
"Kills the line *left* of point, rather than right of it"
   (interactive)
   (setq end (point))
   (beginning-of-line 1)
   (kill-region (point) end))

(defun kill-to-top ()
  "Kill region between point and top of buffer"
  (interactive)
  (kill-region (point-min) (point)))

(defun kill-to-bottom ()
  "Kill region between point and end-of-buffer"
  (interactive)
  (kill-region (point) (point-max)))

(defun bind-macro (key)
"Bind the last macro to a key, locally"
  (interactive "kBind macro to key:")
  (name-last-kbd-macro 'this-macro)
  (local-set-key key 'this-macro))

;;; Emacs Lisp library (non-interactive)
;;; Paul Davis

(defun strip-white-space (string)
  "Return STRING stripped of surrounding white-space"
  (substring string
	     (string-match "[^ 	]" string)
	     (string-match "[ 	]*$" string)))

(defun scar (string)
  "Like car, but for space-separated substrings. Leading and
trailing blank space on the string is ignored"
  (if (string-match "\\(\\s +\\|\n\\)" string)
      (if (= (match-beginning 0) 0)
	  (scar (substring string (match-end 0)))
	(substring string 0 (match-beginning 0)))
    string))

(defun scdr (string)
  "Like cdr, but for space-separated substrings. Leading and trailing
white space on the string is ignored."

  (if (string-match "\\(\\s +\\|\n\\)" string)
      (if (= (match-end 0) (length string))
	  (substring string 0 (match-beginning 0))
	(substring string (match-end 0)))
    ""))

(defun resolve-string (string)
  "Return a list of strings where each element is
an item of STRING. Items are space-separated in STRING."
  (cond ((= (length string) 0) nil)
	(t (cons (scar string) (resolve-string (scdr string))))))

(defun expand-path-names (path)
  (cond ((null path) nil)
	(t (cons (expand-file-name (car path)) 
		 (expand-path-names (cdr path))))))

(defun pop-this-file (path file)
  "Return PATH with FILE as first component."
  (setq new-path (strip-this-file path file))
  (cons file new-path))

(defun strip-this-file (path file)
  "Return PATH without FILE as a component"
  (cond ((null path))
	((string= (car path) file)
	 (cdr path))
	(t
	 (cons (car path) (strip-this-file (cdr path) file)))))

(defun file-type-match (file-name file-type)
"Return true if there is a terminating substring of file-name preceded by a `.'
and that matches file-type"
  (interactive "sFilename: \nsfiletype:")
  (if (eq nil (setq suffx-ndx (string-match "\\." file-name nil)))
      nil
    (equal file-type (substring file-name (+ suffx-ndx 1) nil))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;				NO WARRANTY
;;; 
;;; This software is distributed free of charge and is in the public domain.
;;; Anyone may use, duplicate or modify this program.  Thinking Machines
;;; Corporation does not restrict in any way the use of this software by
;;; anyone.
;;; 
;;; Thinking Machines Corporation provides absolutely no warranty of any kind.
;;; The entire risk as to the quality and performance of this program is with
;;; you.  In no event will Thinking Machines Corporation be liable to you for
;;; damages, including any lost profits, lost monies, or other special,
;;; incidental or consequential damages arising out of the use of this program.
;;;
;;; Thu May  4 1989
;;;
;;; $Header: /tmp_mnt/am/p7/utility/gmacs/f2/RCS/extras.el,v 1.1 88/12/27 17:06:18 fad Exp $

;; Extra little functions which got left out of Emacs.
;; Written by John R. Rose (Rose@THINK.COM).

(provide 'extras)

;;; Table of contents:
;; lazy-switch-to-buffer	switch-to-buffer, with less fuss
;; recursive-edit-once		for command loops, executes on Emacs command
;; file-all-versions, file-previous-version
;;				computation on file version numbers
;; nth-recent-buffer, interesting-buffer-p
;;				access to buffer history
;; switch-to-other-buffer	general buffer history command
;; insert-date			inserts date (and time too)
;; concat-words			like concat, but uses blanks
;; quote-for-shell		knows about Unix shell escapes
;; slashify-line, slashify-region-lines
;;				add continuation chars to ends of lines
;; edit-string			special form applies editor commands to string
;; read-filtered-file-name	upgrade to read-file-name
;; looking-after		reverse version of looking-at
;; forward-atom			like forward-word, skips w and _ syntax
;; current-window-line, window-line-position
;;				ask for relations between point & window line
;; buffer-represented-by-file, file-represented-by-buffer
;;				tells if buffer is synchronized w/ disk
;; extract-file-and-line	parse various forms of "line 10, file foo.c"
;; find-source-code		uses extract-file-and-line, then goes to location;
;;				very useful for following up error messages
;; read-string-no-echo		useful for reading passwords



;; A switch-to-buffer which disturbs the screen less:

(defun lazy-switch-to-buffer (buffer &optional norecord)
  "Just like switch-to-buffer, except it reuses a window if BUFFER is visible."
  (setq buffer (get-buffer buffer))
  (let ((window (get-buffer-window buffer)))
    (if window
	(select-window window)))
  (switch-to-buffer buffer norecord)
  buffer)

;; One-shot recursive edit:

(defun recursive-edit-once (&optional prompt first-char)
  "Effectively does a recursive edit, but only for one command."
  (and first-char (setq unread-command-char first-char))
  (let ((cmd) (flag t))
    (while flag
      (setq cmd (key-binding (read-key-sequence prompt)))
      (if (null cmd)
	  (ding)
	(command-execute cmd))
      (if (null prefix-arg)
	  (setq flag nil)))
    cmd))

;; File version manipulation:

(defun file-all-versions (file)
  "Returns a list of all versions of FILE, in decreasing order.  (FILE is first.)"
  (let ((version-control t) (kept-new-versions 1) (kept-old-versions 0))
    (let ((list (find-backup-file-name file)))
      (setq list (cdr list));; Don't want new backup name.
      (let ((single-backup (and (null list)
				(make-backup-file-name file))))
	(and single-backup
	     (file-exists-p single-backup)
	     (setq list (cons single-backup list))))
      (cons file (nreverse list)))))

(defun file-previous-version (file &optional nth)
  "Returns the filename of the previous version of FILE, or nil.
   Optional arg means get the NTH previous version; default is 1."
  (let* ((base-file (if (backup-file-name-p file) (file-name-sans-versions file) file))
	 (versions (file-all-versions base-file)))
    (while (and versions (not (eql file (car versions)))
		(not (string-equal file (car versions))))
      (setq versions (cdr versions)))
    (nth (or nth 1) versions)))



(defun interesting-file-name-p (file)
  "Is the given file interesting, according to completion-ignored-extensions?"
  (let* ((flen (length file))
	 (fch (aref file (1- flen)))
	 (xptr completion-ignored-extensions))
    (while (and xptr file)
      (let* ((ext (car xptr))
	     (xlen (length ext)))
	(setq xptr (cdr xptr))
	(and (= fch (aref ext (1- xlen)))
	     (< xlen flen)
	     (or (= 1 xlen)
		 (and (= (aref ext (- xlen 2)) (aref file (- flen 2)))
		      (string-equal ext (substring file (- flen xlen)))))
	     (setq file nil)))))
  file)

(defun file-name-good-completions (file dir)
  "Return a list of good completions of file name FILE in directory DIR.
   Filters out files according to completion-ignored-extensions."
  (let* ((res (cons nil (file-name-all-completions file dir)))
	 (ptr res))
    (while (cdr ptr)
      (if (interesting-file-name-p (car (cdr ptr)))
	  (setq ptr (cdr ptr))
	(rplacd ptr (cdr (cdr ptr)))))
    (cdr res)))
	


;;; Buffer "history".

(defun switch-to-other-buffer  (&optional arg)
  "Switches to the other buffer, as determined by (other-buffer).
   With a non-null ARG (numeric prefix), switches to the nth buffer
   (1-origin) in the history.  As with ZMACS, 0 and 1 are special cases:
   0 means display a buffer list, while 1 means bury current buffer."
  (interactive "P")
  (cond
   ((null arg)
    (switch-to-buffer (other-buffer)))
   ((= 0 (setq arg
	       (prefix-numeric-value arg))) ;; In case C-U, etc.
    (electric-buffer-list nil))
   ((= 1 arg)
    (bury-buffer))
   (t (switch-to-buffer
       (or (nth-recent-buffer (1- arg))
	   (error "Not that many buffers."))))
   ))

(defun nth-recent-buffer (arg)
  ;(nth (1- arg) (interesting-buffer-list))
  (let ((buf nil)
	(bufs (buffer-list)))
    (while (and bufs (not (minusp
			   (prog1 arg (setq arg (1- arg))))))
      (while (and bufs
		  (not (interesting-buffer-p
			(prog1
			    (setq buf (car bufs))
			  (setq bufs (cdr bufs))))))))
    buf
    ))

(defun interesting-buffer-p (buffer)
  (setq buffer (buffer-name (get-buffer buffer)))
  (or (zerop (length buffer))
      (not (eql ?\  (elt buffer 0)))))



(defun insert-date (and-time)
  "Insert today's date at point in the buffer.  If argument is non-null, use time too."
  ;; E.g., "Tue Sep 15 18:29:44 1987"
  (interactive "P")
  (let ((time (current-time-string)))
    (or and-time (setq time (concat (substring time 0 (string-match " *[0-9]+:.*" time))
				    (substring time (string-match " *[0-9]*$" time))
				    )))
    (insert time)))



;; For dealing with Unix shell:

(defun concat-words (&rest args)
  "Works like concat, but inserts spaces between the ARGS if necessary.
   Used to create Unix command lines."
  (setq args (apply 'list args))
  (let ((ptr args))
    (while ptr
      (let* ((word1 (car ptr))
	     (len1 (length word1))
	     (word2 (car (cdr ptr)))
	     (len2 (length word2)))
	(or (null (cdr ptr))
	    (and (plusp len1) (= ?  (elt word1 (1- len1))))
	    (and (plusp len2) (= ?  (elt word2 0)))
	    (rplacd ptr (setq ptr (cons " " (cdr ptr)))))
	(setq ptr (cdr ptr)))))
  (apply 'concat args))

(defvar shell-meta-characters " \t\n[]*?`'\"\\~$!#^&|;()<>")

(defun quote-for-shell (string)
  "If STRING has any Unix shell meta-characters in it, quote them."
  (and (stringp shell-meta-characters)
       (setq shell-meta-characters (append shell-meta-characters ())))
  (or (listp shell-meta-characters)
      (error "shell-meta-characters: not a list"))
  (let ((i (length string))
	(bad 0) (very-bad nil) ch)
    (while (and (plusp i) (not very-bad))
      (setq i (1- i)
	    ch (elt string i))
      (when (memq ch shell-meta-characters)
	(setq bad (1+ bad))
	(when (memq ch '(?\' ?\!)) (setq very-bad t))))
    (cond
     ((zerop bad) string)
     ((or very-bad (< bad 2))
      (setq string (append string ()))
      (let ((ptr string) ch)
	(while ptr
	  (setq ch (car ptr))
	  (if (not (memq ch shell-meta-characters))
	      (setq ptr (cdr ptr))
	    (rplaca ptr ?\\)
	    (rplacd ptr (cons ch (cdr ptr)))
	    (setq ptr (cdr (cdr ptr))))))
      (concat string))
     (t (concat "'" string "'")))))



(defvar slashify-region-column 72 "*Column at which slashification takes place.")
(defvar slashify-region-string "\\" "*String with which slashified lines end.")
(make-variable-buffer-local 'slashify-region-column)
(make-variable-buffer-local 'slashify-region-string)

(defun set-slashification ()
  (interactive)
  "Sets up slashification parameters based on the current column
   and the string between point and end of line."
  (save-excursion
    (setq slashify-region-column (current-column)
	  slashify-region-string (buffer-substring (point) (progn (end-of-line) (point)))))
  (message "column set to %d, string set to %s (local to this buffer)"
	   slashify-region-column slashify-region-string))

(defun slashify-region-lines (&optional start end column string)
  "Fixes up slashification of lines in the region START to END.
   Interactive prefix gives non-default COLUMN to adjust to;
   negative or non-numeric COLUMN means delete slashification.
   STRING gives a non-default suffix to add to the line."
  (interactive "r\nP")
  ;; Noninteractive args:
  (or start (setq start (point)))
  (or end (setq end (mark)))
  (or (< start end) (setq start (prog1 end (setq end start))))
  (or column (setq column slashify-region-column))
  (or string (setq string slashify-region-string))
  (let ((slashify-region-string string)
	(slashify-region-column column))
    ;; ^^^ Args for slashify-line.
    (let ((indent-line-function 'slashify-line)
	  (indent-region-function nil))
      (indent-region start end nil))
    ))
(defun slashify-line (&optional column string)
  "Fix up slashification on one line.  Prefix is as for slashify-region-lines."
  (interactive "P")
  (or column (setq column slashify-region-column))
  (or string (setq string slashify-region-string))
  (or (and (integerp column) (>= column 0)) (setq column nil))
  (let (col0 colM (strlen (length string)))
    (beginning-of-line)
    (setq col0 (point))
    (end-of-line)
    (setq colM (point))
    (and (> strlen 0) (<= strlen (- colM col0))
	 (string-equal string (buffer-substring (- colM strlen) colM))
	 (delete-char (- strlen)))
    (delete-horizontal-space)
    (if (null column)
	();; We're done.
      (indent-to column 1)
      (insert string))
    ))



;; String editing facility:

(defvar in-funcall-edit-string nil)
(defun funcall-edit-string (string func &rest args)
  (save-excursion
    (if in-funcall-edit-string
	(switch-to-buffer
	 (generate-new-buffer " *Edit String Recursive*") t)
      (switch-to-buffer ;;set-buffer gets spectacular core dump
       (get-buffer-create " *Edit String Workspace*") t)
      (setq buffer-read-only nil)
      (fundamental-mode))
    (delete-region (point-max) (point-min))
    (insert-string string)
    (goto-char (point-min))
    (let ((in-funcall-edit-string t))
      (apply func args))
    (prog1
	(buffer-string)
      (if in-funcall-edit-string
	  (kill-buffer (current-buffer))))
    ))

(put 'funcall-edit-string 'lisp-indent-hook 1)
(put 'edit-string 'lisp-indent-hook 1)

(defmacro edit-string (string &rest body)
  "Takes STRING, puts it into a new buffer, and runs the body code.
   Returns the resultant edited string."
  (list 'funcall-edit-string string
	(list 'quote (nconc (list 'lambda ()) body))))



;; File system access functions:

(defun file-has-changed-p (file-props)
  "Tells whether a file has changed since the last call to this function.
   FILE-PROPS is a cons of a file name string and the result of
   file-attributes on that file.  The cdr is always updated in place
   to the current attributes."
  (let ((old-attr (cdr file-props))
	(new-attr (file-attributes (car file-props))))
    (if (equal (nthcdr 5 old-attr) (nthcdr 5 new-attr))
	nil ;; No change.
      (rplacd file-props new-attr)
      (car file-props))
    ))

;; Nifty abstract filename reader.
;; By R. Thau & J. Rose.

(defun file-name-completion-table (&optional directory full match field)
  "Returns alist useful for completing reads of names of files in DIRECTORY.
DIRECTORY is optional, and defaults to default-directory.

The alist is useful with the function completing-read (which see).

The cdrs of the cons cells in the alist are the file names in DIRECTORY.
As with directory-files, optional argument FULL makes these names absolute.
The optional third argument MATCH is a regexp which filters the names put
in the list.

The cars of the alist are what completing-read expects to read.
By default they are just the base file names, but if the optional fourth
argument FIELD is specified, it is the number of a subexpression of MATCH;
only thse subexpressions are put in the cars of the alist.

Thus (file-name-completion-table dir t \"^\\\\(.*\\\\)\\\\.c\" 1) returns a
table suitable for reading the name of a C source file in \"dir\", hiding
the \".c\" extension from the user.

MATCH can also be a functional predicate to apply to each possible file name;
interesting-file-name-p is a useful choice.  FIELD can be used in this
case too; if the predicate succeeds it must have performed a regexp
operation such as string-match.
"
  (let* ((filter (if (not (stringp match)) (prog1 match (setq match nil))))
	 (files (directory-files directory full match))
	 (alist nil))
    (while files
      (let* ((fname (car files))
	     (key (file-name-nondirectory fname)))
	(setq files (cdr files))
	(if field
	    (progn (if filter
		       (funcall filter key)
		     (string-match match key))
		   (setq key (substring key
					(match-beginning field)
					(match-end field)))))
	(setq alist (cons (cons key fname) alist))))
    (nreverse alist)))

(defun read-filtered-file-name
  (prompt &optional directory default match field cache)
  "Completing read of the name of a file; works like read-file-name.

The first three arguments are the same as for read-file-name:
PROMPT is used as a prompt for the read, DIRECTORY is the directory,
and DEFAULT is returned if a null string is entered.
The 'MUSTMATCH' argument of read-file is not taken,
because all responses (except a non-null DEFAULT) are
required to match existing files.

The fourth and fifth arguments MATCH and FIELD further restrict
the possible inputs; MATCH is usually a regexp, and a non-null FIELD
is the number of a subexpression within the regexp.
See file-name-completion-table for further details.

The ambient value of insert-default-directory determines whether
a absolute or relative file name is returned.  This is compatible
with read-file-name, although you never see the directory.

The sixth argument CACHE, if non-null, is a symbol whose value
cell is used to reduce scans of the directory.  If the directory
has changed, file-name-completion-table must be called."
  ;; Default directory:
  (or directory (setq directory default-directory))
  ;; Normalize cache to ((dir . props) . alist):
  (and cache
       (not (and (boundp cache) (symbol-value cache)))
       (set cache (list (list directory))))
  (setq cache (symbol-value cache));; Mutate list structure hereafter.
  (let ((alist
	 (and cache
	      (not (file-has-changed-p (car cache)))
	      (cdr cache))))
    (if (null alist)
	(progn
	  (setq alist (file-name-completion-table
		       directory insert-default-directory match field))
	  (and cache (rplacd cache alist))))
    ;; Got the alist, now use it.
    (let* ((key (completing-read prompt alist nil t))
	   (item (if (plusp (length key))
		     (assoc key alist)
		   (and default (cons nil default)))))
      (or item (error "No default."))
      (cdr item)
      )))



;; Reverse version of looking-at, very good for parsing prompts:
(defun looking-after (pat)
  "t if text just before point matches regular expression PAT.
   Just the reverse of looking-at.  Sets up the match markers, of course.
   The match chosen differs from that of re-search-backward,
   in that it is the earliest rather than the latest possible match,
   and the match must end at point.  (The search is carried out such
   that matches extending past point are never considered.)"
  ;; Just try to do this more efficiently!
  (let ((opoint (point))
	(npoint nil))
    (save-restriction
      ;; Can't overflow matches past initial point:
      (narrow-to-region (point-min) opoint)
      (re-search-backward pat nil t)
      (while (and (looking-at pat)
		  (= opoint (match-end 0))
		  (prog1 t (setq npoint (point)))
		  (not (bobp)))
	(forward-char -1))
      ;; Make sure the match was anchored:
      (and npoint
	   (or (= npoint (match-beginning 0))
	       (progn (goto-char npoint)
		      (looking-at pat))))
      (goto-char opoint)
      (not (null npoint))
      )))



(defun forward-atom (&optional arg)
  "Like forward-word, but also skips over chars with symbol syntax."
  (interactive "p")
  (or arg (setq arg 1))
  (let ((n arg)
	ch flag)
    (while (and (plusp n)
		(not (eobp)))
      (setq flag t)
      (while flag
	(setq ch (char-after (point)))
	(setq ch (if ch (char-syntax ch) ?w))
	(cond ((= ch ?_) (setq flag nil))
	      ((= ch ?w) (setq flag nil))
	      (t (forward-char 1))))
      ;; Looking at an atom.  Move over it.
      (setq flag t)
      (while flag
	(setq ch (char-after (point)))
	(setq ch (if ch (char-syntax ch) -1))
	(cond ((= ch ?_) (forward-char 1))
	      ((= ch ?w) (forward-word 1))
	      (t (setq flag nil))))
      (setq n (1- n)))
    (while (and (minusp n)
		(not (bobp)))
      (setq flag t)
      (while flag
	(setq ch (char-after (1- (point))))
	(setq ch (if ch (char-syntax ch) ?w))
	(cond ((= ch ?_) (setq flag nil))
	      ((= ch ?w) (setq flag nil))
	      (t (forward-char -1))))
      ;; Looking at an atom.  Move over it.
      (setq flag t)
      (while flag
	(setq ch (char-after (1- (point))))
	(setq ch (if ch (char-syntax ch) -1))
	(cond ((= ch ?_) (forward-char -1))
	      ((= ch ?w) (forward-word -1))
	      (t (setq flag nil))))
      (setq n (1+ n)))
    (zerop n)
    ))



;; Window inquiries, and one more movement function:

(defun current-window-line (&optional window fixup-p)
  "Returns the window line that point is on in WINDOW.  WINDOW is optional.
   If the window's point is currently invisible, returns nil, or, if
   optional FIXUP-P is non-nil, scrolls the window to accommodate point.
   This function, as applied to (point), is inverse to window-line-position.
   (recenter (current-window-line)) is a no-op, and
   (move-to-window-line (current-window-line)) amounts to (beginning-of-line)."
  (setq window (get-window window))
  (let ((old-window (selected-window))
	(res 0)
	start opoint)
    (if (eq window old-window)
	(setq old-window nil)
      (select-window window))
    (setq opoint (point)
	  start (window-start))
    (vertical-motion 0)
    (cond
     ((pos-visible-in-window-p (point))
      (while (> (point) start)
	(vertical-motion -1)
	(setq res (1+ res))))
     ((not fixup-p)
      (setq res nil))
     ((< (point) start)
      (set-window-start window (point)))
     (t
      (setq res (max 0 (- (window-height) 2)))
      (vertical-motion (- res))
      (set-window-start window (point))))
    (goto-char opoint)
    (and old-window (select-window old-window))
    res))

(defun window-line-position (line &optional window)
  "Returns the position (as a buffer point) of the start of the given
   LINE in the current window, or WINDOW if given.
   As with the commands recenter and move-to-window-line,
   LINE is zero-origin, and may be negative."
  (cond
   ((zerop line)
    (window-start window))
   ((null window)
    (let ((height (1- (window-height))))
      (if (minusp line) (setq line (+ line height)))
      ;; Error check:
      (if (minusp line) (setq line 0))
      (if (>= line height) (setq line (1- height)))
      (save-excursion
	(goto-char (window-start))
	(vertical-motion line)
	(point))
      ))
   (t (let ((old-window (selected-window)))
	(if (eq old-window window)
	    (window-line-position line)
	  (unwind-protect
	      (progn
		(select-window window)
		(window-line-position line))
	    (select-window old-window)))
	))
   ))

(defun get-window (window)
  "Returns the window associated with WINDOW, which is
   either a window, a buffer, or a buffer name."
  (cond
   ((null window)
    (selected-window))
   ((windowp window)
    window)
   (t (let ((buffer (get-buffer window))
	    (this-window (selected-window)))
	(if (eq buffer (window-buffer this-window))
	    this-window
	  (or (and buffer
		   (get-buffer-window buffer))
	      (error "No such window: %s" (or buffer window)))
	  )))))


(defmacro with-window-selected (window &rest body)
  "Causes a switch to the given WINDOW, the execution of BODY,
   and the a return to the original window."
  (list
   'let
   (list '(before-with-window-selected (selected-window))
	 (list 'inside-with-window-selected window))
   (list
    'unwind-protect
    (nconc
     (list
      'progn
      '(or (null inside-with-window-selected)
	   (eq before-with-window-selected inside-with-window-selected)
	   (select-window inside-with-window-selected)))
     body)
    '(and (window-start before-with-window-selected)
	  (select-window before-with-window-selected))
    )))



(defun recenter-window (line window)
  "Perform a recenter command with argument LINE in the given WINDOW."
  (if (or (null window)
	  (eq window (selected-window)))
      (recenter line)
    (let ((old-window (selected-window)))
      (unwind-protect
	  (progn
	    (select-window window)
	    (recenter line))
	(select-window old-window)))))



;; Patch to allow recursive walk of keymap:

(defun substitute-key-definition (olddef newdef keymap &optional recursivep)
  "Replace OLDDEF with NEWDEF for any keys in KEYMAP now defined as OLDDEF.
In other words, OLDDEF is replaced with NEWDEF where ever it appears.
Keymaps for prefix chars inside KEYMAP are only traversed if the optional
flag RECURSIVEP is t."
  (if (arrayp keymap)
      (let ((len (length keymap))
	    (i 0)
	    def)
	(while (< i len)
	  (setq def (aref keymap i))
	  (cond ((eq def olddef)
		 (aset keymap i newdef))
		((and recursivep
		      (keymapp def))
		 (substitute-key-definition olddef newdef def recursivep)))
	  (setq i (1+ i))))
    (while keymap
      (setq def (cdr-safe (car-safe keymap)))
      (cond ((eq def olddef)
	     (setcdr (car keymap) newdef))
		((and recursivep
		      (keymapp def))
		 (substitute-key-definition olddef newdef def recursivep)))
      (setq keymap (cdr keymap)))))



(defun buffer-represented-by-file (buffer)
  "Returns a file, if any, which holds the current contents of BUFFER."
  (let ((file (buffer-file-name buffer)))
    (and file (file-exists-p file) (not (file-directory-p file))
	 (verify-visited-file-modtime buffer)
	 (not (buffer-modified-p buffer))
	 file)))

(defun file-represented-by-buffer (file)
  "Returns a buffer, if any, which holds the current contents of FILE."
  (let ((buffer (get-file-buffer file)))
    (and buffer (buffer-represented-by-file buffer) buffer)))



;; Suggested binding:      
;;(global-set-key "\C-c\C-f" 'find-source-code)

;; Extracting file locations from the buffer.
;; To do:
;; *  Allow a "try-harder" switch, which uses weaker regexps.
;; *  Integrate with after-file-name in shell-plus.el,
;;    and compilation-grab-filename in compile.el.
;; *  Parse out the diagnostic as a third field.

(defconst file-location-patterns
  '(;; cpp: file.c: 10:
    ("\\(^\\|[ \t]\\)\\([^ \t\n]+\\): *\\([0-9]+\\):" cpp 2 3)
    ;; cpp: #line 10 "file.c"
    ("#\\(line\\)? *\\([0-9]+\\) *\"\\([^\"\n]+\\)\"" cpp 3 2)
    ;; cc: "file.c", line 10
    ("\"\\([^\"\n]+\\)\", line +\\([0-9]+\\)" cc 1 2)
    ;; f77: line 10 of file.c
    ("line +\\([0-9]+\\) +of +\\([^ \t\n]+\\)" f77 2 1)
    ;; saber: "file.c:10"
    ("\"\\([^\"\n]+\\):\\([0-9]+\\)\"" saber 1 2)
    ;; shell: file.c: syntax error at line 10
    ("\\(^\\|[ \t]\\)\\([^ \t\n:]+\\):.*line +\\([0-9]+\\)" sh 2 3)
    ;; lint: : file.c(10)
    (": *\\([^ \t\n)]+\\) *(\\([0-9]+\\))" lint 1 2)
    ;; lint: file.c(10) :
    ("\\(^\\|[ \t]\\)\\([^ \t\n)]+\\) *(\\([0-9]+\\)) *:" lint 2 3)
    ;; lint: ( file.c(10) )
    ("( +\\([^ \t\n)]+\\) *(\\([0-9]+\\)) +)" lint 1 2)
    ;; troff: `file.c', line 10
    ("[\"`']\\([^\"`'\n]+\\)[\"`'], line +\\([0-9]+\\)" troff 1 2)
    ;; ri: "file.c" 10:
    ("\"\\([^\"\n]+\\)\" *\\([0-9]+\\):" ri 1 2) ;;Never heard of ri.
    ;; mod: File file.c, line 10
    ("[Ff]ile +\\([^ \t\n]+\\), line +\\([0-9]+\\)" mod 1 2)
    ;; ksh: file.c[10] :
    ("\\(^\\|[ \t]\\)\\([^ \t\n)]+\\) *\\[\\([0-9]+\\)\\] *:" ksh 2 3)
    )
  "Series of regexps matching file number locations.
   Each list entry is a 4-list of a regexp, a symbol, and two numbers.
   The numbers name regexp fields which will hold the file and line number.
   The symbol is returned unexamined from (extract-file-and-line);
   it should be a guess at who produced the message, e.g., 'cc'.
   These regexps are based on the one in compile.el (although the
   pattern containing 'of' must also contain 'line'),
   on the Unix filters 'error' and 'fwarn'.
   The patterns are known to recognize errors from the following
   Un*x language processors:
     cpp, cc, lex, f77, Saber C, sh (Bourne), lint, mod
   The following language processors do not incorporate file names
   in every error message, and so are more difficult to accomodate:
     yacc, pc, csh
   ")

(defun extract-file-and-line (&optional where)
  "Returns a list of file and line number, obtained by parsing
   the current buffer around WHERE (defaults to current point).
   Uses file-location-patterns to parse with.  WHERE must either
   be inside the location or precede it on the same line.
   WHERE can also be a string, in which case it is parsed.
   The third element of the list is a symbol recording a guess
   as to what produced the message, and the fourth and fifth
   elements are positions at which the file and line were
   found to begin.  Of course, NIL means failure."
  (or where (setq where (point)))
  (if (stringp where)
      (save-excursion
	(set-buffer (get-buffer-create "*Extract File and Line*"))
	(delete-region (point-min) (point-max))
	(insert-string where)
	(extract-file-and-line 0))
    (let ((ptr file-location-patterns)
	  (found nil)
	  pat)
      (while (and ptr (not found))
	(setq pat (car ptr) ptr (cdr ptr))
	(and (or (and (> where (point-min))
		      (save-excursion
			(goto-char where)
			(end-of-line)
			(re-search-backward
			 (car pat) (save-excursion
				     (beginning-of-line) (point)) t))
		      (>= (match-end 0) where))
		 (save-excursion
		   (goto-char where)
		   (re-search-forward
		    (car pat) (save-excursion (end-of-line) (point)) t)))
	     (setq found t)))
      (and found
	   (let ((lang (nth 1 pat))
		 (f1 (nth 2 pat))
		 (f2 (nth 3 pat)))
	     (list
	      (buffer-substring (match-beginning f1) (match-end f1))
	      (string-to-int
	       (buffer-substring (match-beginning f2) (match-end f2)))
	      lang
	      (match-beginning f1)
	      (match-beginning f2)
	      )))
      )))

(defun find-source-code ()
  (interactive)
  (let ((res (or (extract-file-and-line)
		 (error "Not sitting on a file location."))))
    (let ((file (nth 0 res))
	  (line (nth 1 res)))
      (find-file-other-window file)
      (goto-line line)
      (message "line %d of %s" line file)
      )))



;; Password-reader.  Amusing keystroke echoing properties.

(defvar read-string-no-echo-default-echo "."
  "*String used to echo keystrokes read by #'read-string-no-echo.
   Usually a sober \".\", but try \"xyzzy\".")

(defun read-string-no-echo (prompt &optional echo-string end-chars)
  "Issue PROMPT and read a string, without echoing it in the process.
   Optional ECHO-STRING is a string to using in echoing keystrokes;
   it defaults to read-string-no-echo-default-echo.
   Optional END-CHARS is a list of characters to terminate the string;
   the default is CR and LF. Some rubout and help processing is also done."
  (or end-chars (setq end-chars '(?\n ?\r)))
  (if (zerop (length echo-string))
      (setq echo-string read-string-no-echo-default-echo))
  (let ((res nil) (done nil))
    (while (not done)
      (message "%s%s%s" prompt
	       (apply 'concat (make-list (/ (length res) (length echo-string))
					 echo-string))
	       (substring echo-string 0 (% (length res) (length echo-string))))
      (let ((ch (let ((cursor-in-echo-area t)) (read-char))))
	;; (Could use the isearch chars here.)
	(cond ((memq ch end-chars) (setq done t))
	      ((= ch ?\C-?) (setq res (cdr res)))
	      ((= ch ?\C-G) (setq res nil done t))
	      ((= ch help-char)
	       (message "%s [CR to end, C-G to abort, DEL to rubout]" prompt)
	       (sit-for 1))
	      (t (setq res (cons ch res))))))
    (concat (nreverse res))))

